package PIANOS.generator;

import PIANOS.datastructures.*;
import PIANOS.exceptions.*;
import PIANOS.io.*;

import java.util.*;
import java.io.*;

/**
 * This class is used to generate the Fortran module "proposal".
 */
public class Proposal {

    /**
     * Generates and writes the Fortran module proposal.f90. 
     *
     * @param  model The model this simulation is related to
     */
    public static void generateProposal(ComputationalModel model) throws InvalidProposalException, 
    IllegalParametersException, MissingFunctionException, IOException {
        FortranWriter out = new FortranWriter("proposal.f90");
        
        ArrayList <String> result = new ArrayList<String>();
        
        result.add("MODULE proposal");
        result.add("USE user_dist");
        result.add("IMPLICIT NONE");
        result.add("PRIVATE");
        result.add("INTERFACE generate");
        result.add("MODULE PROCEDURE generate_int, generate_real");
        result.add("END INTERFACE");
        result.add("PUBLIC :: generate ");
        result.add("CONTAINS");
    
    
        result.add("SUBROUTINE generate_int(name, buffer)");
        result.add("IMPLICIT NONE");
        result.add("CHARACTER(LEN=*), INTENT(IN) :: name");
        result.add("INTEGER, DIMENSION(:), INTENT(INOUT) :: buffer");
        result.add("");
        result.add("INTEGER :: k, ifail");
        result.add("INTEGER, PARAMETER :: dp = SELECTED_REAL_KIND(8, 0)");
        result.add("REAL(KIND=dp), DIMENSION(:), ALLOCATABLE :: binomial_reference");
        result.add("REAL(KIND=dp), DIMENSION(:), ALLOCATABLE :: poisson_reference");
        result.add("");
        result.add("EXTERNAL G05ECF");
        result.add("INTEGER G05EYF");
        result.add("EXTERNAL G05EYF");
        result.add("INTEGER G05DYF");
        result.add("EXTERNAL G05DYF");
        result.add("EXTERNAL G05FAF");
        result.add("EXTERNAL G05EDF");
        result.add("EXTERNAL G05FEF");
        result.add("");
        result.add("SELECT CASE (name)");
	
        LinkedList<Variable> allVariables = new LinkedList<Variable>();
        allVariables.addAll(model.getVariableList());

	    LinkedList<Entity> entities = model.getEntityList();

        for (Entity e : entities ) {
	        allVariables.addAll(e.getVariableList());
	    }
	
	    for (Variable v : allVariables) {
            if (v.isData() && v.getMissingValueCount() == 0) {
                continue;
	        }
            if (v.isFunctional()) {
                continue;
            }
            if (v.isInteger()) {
                result.add("CASE('" + v.getName() + "')");
                    
                int paramCount = v.getProposal().getNumberOfParameters();
                String[] parameters = new String[paramCount+1];
                    
                for (int i=0; i<paramCount; i++) {
                    Object par = v.getProposal().getParameter(i);

                    if (par instanceof Integer) {
                    parameters[i] = ((Integer)par).toString(); 
                    }
                    else if (par instanceof Double) {
                    parameters[i] = ((Double)par).toString() + "_dp";
                    }
                    else throw new InvalidProposalException("The proposal distribution of " 
                                    + v.getName() + " contains reference parameters.");
                }
                parameters[paramCount] ="buffer";

                result.addAll(v.getProposal().getGenCode(parameters));
            }
    	}

        LinkedList<Variable> variables = model.getVariableList();
        
        // DEBUG
        /*
        System.out.println("GENERATOR DEBUG Proposal.java: Global variable list after variable cases");
        for (Variable v: variables)
            System.out.println(v.getName());
        */

        result.add("END SELECT");

        result.add("END SUBROUTINE generate_int");
        result.add("");
        result.add("SUBROUTINE generate_real(name, buffer)");
        result.add("IMPLICIT NONE");
        result.add("CHARACTER(LEN=*), INTENT(IN) :: name");
        result.add("REAL(KIND=SELECTED_REAL_KIND(8, 0)), DIMENSION(:), INTENT(INOUT) :: buffer");
        result.add("INTEGER, PARAMETER :: dp = SELECTED_REAL_KIND(8, 0)");
        result.add("");
        result.add("INTEGER :: ifail, k");
        result.add("");
        result.add("EXTERNAL G05ECF");
        result.add("INTEGER G05EYF");
        result.add("EXTERNAL G05EYF");
        result.add("INTEGER G05DYF");
        result.add("EXTERNAL G05DYF");
        result.add("EXTERNAL G05FAF");
        result.add("EXTERNAL G05EDF");
        result.add("EXTERNAL G05FEF");
        result.add("");
        result.add("SELECT CASE (name)");

        for (Variable v : allVariables) {
            if (v.isData() && v.getMissingValueCount() == 0) {
                continue;
            }
            if (v.isFunctional()) {
                continue;
            }
            if (v.isInteger() == false) {
                result.add("CASE('" + v.getName() + "')");
                
                int paramCount = v.getProposal().getNumberOfParameters();
                String[] parameters = new String[paramCount + 1];

                for (int i=0; i<paramCount; i++) {
                    Object par = v.getProposal().getParameter(i);

                    if (par instanceof Integer) {
                        parameters[i] = ((Integer)par).toString(); 
                    }
                    else if (par instanceof Double) {
                        parameters[i] = ((Double)par).toString() + "_dp";
                    }
                    else throw new InvalidProposalException("The proposal distribution of " 
                                        + v.getName() + " contains reference parameters.");
                }

                parameters[paramCount] = "buffer";
                
                result.addAll(v.getProposal().getGenCode(parameters));
            }
            
        }
        result.add("END SELECT");
	

        result.add("END SUBROUTINE generate_real");
        result.add("END MODULE proposal");

        out.write(result);
    }

}
