package PIANOS.generator;

import PIANOS.datastructures.*;
import PIANOS.exceptions.*;
import PIANOS.io.*;

import java.util.*;
import java.io.*;

/**
 * Generates calculations needed when generating the Fortarn module "main".
 */
public class Acceptation {
    
    private static ArrayList<Variable> topologicalList;
    
    /**
     * Sets the topological variable list that is used when calculating the correct
     * new-value-generation order for functional parameters. This method must be called before any
     * other methods of this class are called.
     *
     * @param  list A list of all variables in the current model in topolocial order.
     */
    public static void setTopologicalList(ArrayList<Variable> list) {
        // This method has to be called before any other methods!
        topologicalList = new ArrayList<Variable>(list);
    }
    
    
    /**
     * Generates code that fetches a proposed new value for a parameter from its buffer or if the buffer
     * is empty, calls the subroutine "generate".
     *
     * @param  variable A parameter for which the new value code is generated.We assume that it
     * is indexed by "one_dim(1)" if it is global, by "one_dim(i)" if it is one-dimensional and by
     * "two_dim(i, j)" if it is two-dimensional.
     */
    public static ArrayList<String> generateNewValueCode(Variable variable) {
        ArrayList<String> result=new ArrayList<String>();
        
        assert variable.isFunctional() == false;
        assert variable.isData() == false || variable.getMissingValueCount() > 0;
        
        String name = variable.getName();
        result.add("IF (" + name + " % buffer_index > SIZE(" + name + " % buffer) .OR. " + name + " % buffer_index < 1) THEN");
        result.add("CALL generate('" + name + "', " + name + " % buffer)");
        result.add(name + " % buffer_index = 1");
        result.add("END IF");
        result.add("");
        
        String indexing="";
        if (variable.getEntity() == null) {
            indexing = "one_dim(1)";
        } else if (variable.getEntity().isMatrix() == false) {
            indexing = "one_dim(i)";
        } else {
            indexing = "two_dim(i, j)";
        }
        
        assert variable.getStrategy() != null;
        if (variable.getStrategy().equals("random walk")) {
            result.add(name + " % " + indexing + " % new_value = " + name + " % " + indexing + " % value + "
                    + name + " % buffer(" + name + " % buffer_index)");
        } else {
            result.add(name + " % " + indexing + " % new_value = " + name + " % buffer(" + name + " % buffer_index)");
        }
        
        result.add(name + " % buffer_index = " + name +" % buffer_index + 1");
        
        return result;
    }
    
    
    /**
     * Generates code that decides whether the new value is accepted and makes the necessary
     * changes: updates the value of the current parameter and all functional parameters depending
     * on it.
     *
     * @param  variable A parameter for which the acceptation code is generated. We assume that it
     * is indexed by "one_dim(1)" if it is global, by "one_dim(i)" if it is one-dimensional and by
     * "two_dim(i, j)" if it is two-dimensional.
     */
    public static ArrayList<String> generateAcceptationCode(Variable variable, String strategy) throws InvalidModelException{
        
        assert variable.isFunctional() == false;
        assert variable.isData() == false || variable.getMissingValueCount() > 0;
        
        ArrayList<String> result=new ArrayList<String>();
        
        String name = variable.getName();
        
        String indexing="";
        if (variable.getEntity() == null) {
            indexing = "one_dim(1)";
        } else if (variable.getEntity().isMatrix() == false) {
            indexing = "one_dim(i)";
        } else {
            indexing = "two_dim(i, j)";
        }
        
        result.add("IF (p_acc > prob(prob_index)) THEN");
        result.add(name + " % " + indexing + " % value = " + name + " % " + indexing + " % new_value");
        result.add(name + " % " + indexing + " % successful_changes = " + name +" % " + indexing +
                " % successful_changes + 1");
        result.add("");
        if (strategy.equals("random")){
            result.add("IF ("+ name +" % "+ indexing +" % successful_changes == "+
                    name +" % "+ indexing +" % update_count) THEN");
            result.add("parameters_achieved = parameters_achieved + 1");
            result.add("END IF");
        }
        // here we must update all the functionals depending on variable
        // like blah % value = blah % new value
        // problem: spatial case! (we need to _calculate_ the new value)
        
        // 1) functional, non-spatial
        ArrayList<Variable> allAffected = new ArrayList<Variable>(variable.getAffectsList());
        for (int i=0; i<allAffected.size(); i++) {
            Variable depending = allAffected.get(i);
            
            if (!depending.isFunctional())
                continue;
            if(depending.isSpatial())
                continue;
            
            allAffected.addAll(depending.getAffectsList());
            
            LinkedList<Variable> depdeps = depending.getDependsList();
            
            boolean spatialDependencies = false;
            for (Variable depdep : depdeps){
                if (depdep.isSpatial() && allAffected.contains(depdep))
                    spatialDependencies = true;
            }
            
            // code for depending...
            if (spatialDependencies == false) {
                result.addAll(acceptationAffectedOf(variable, depending));
            }
            
        }
        
        // 2) functional, spatial
        // Here we update the spatial parameters of _neighbour_squares_ by using the new value of
        // the current variable of the current square.
        
        allAffected = new ArrayList<Variable>(variable.getAffectsList());
        for (int i=0; i<allAffected.size(); i++) {
            Variable depending = allAffected.get(i);
            
            if (!depending.isFunctional())
                continue;
            LinkedList<Variable> depdeps = depending.getDependsList();
            
            boolean spatialDependencies = false;
            for (Variable depdep : depdeps){
                if (depdep.isSpatial() && allAffected.contains(depdep))
                    spatialDependencies = true;
            }
            
            /* now, if this variable (depending) has spatial dependencies:
             * 1. depending is not spatial. (spatial variables have only one non-spatial dep)
             * 2. depending has not been updated yet in the previous loop.
             * 3. the spatial dependencies of depending have their new values calculated
             *    previously in this loop.
             * So we can safely update it as we did before
             * (and the path of other functionals leading from it)
             * TROUBLE. If this depending is affected by a global, we must add loops and
             * we shouldn't even be doing this neighbour-thing.
             * How to recognize that the new value for depending has been already calculated
             * while considering this global? ... Maybe we should look at the dependsList of depending.
             * not really. Just checking that variable.getEntity() == null should suffice.
             * IS IT ALWAYS like this? can we always skip the neighbouring process with variables
             * that are global?
             */
            
            
            if (spatialDependencies){
                if (variable.getEntity() == null){
                    result.addAll(acceptationAffectedOf(variable, depending)); 
                } else{
                    boolean spatialY;
                    if (depending.getEntity().isMatrix() == false){
                        indexing = " % one_dim(neighbour)";
                        spatialY = true;
                    } else{
                        if (depending.getEntity().getXCoordinate().isSpatial()){
                            indexing = " % two_dim(i, neighbour)";
                            spatialY = false;
                        } else{
                            indexing = " % two_dim(neighbour, j)";
                            spatialY = true;
                        }
                    }   

                    if (spatialY){
                        result.add("DO k = 1, spatial(i, 1)");
                        result.add("neighbour = spatial(i,  k +1)");
                    } else{
                        result.add("DO k = 1, spatial(j, 1)");
                        result.add("neighbour = spatial(j,  k +1)");
                    }
                        
                    result.addAll(generateSpatialDeps(depending, indexing, spatialY));
                    result.add("END DO");
                        
                        
                 
                } // end of not global variable
            } // end of spatial deps
            if(!depending.isSpatial()) // updated already
                continue;
            
            for (Variable affected: depending.getAffectsList()){
                if (!allAffected.contains(affected))
                    allAffected.add(affected);
            }
            
            // adding further forks...
            // Spatial variables are direct, but the functionals
            // that depend on them have not been updated yet.
            
            /* ok, now update the spatial variable.
             * it's either COUNT(&variable)
             * or SUM(&variable)
             * (since it depends on variable)
             */
            String depname = depending.getName();
            
            String[] depEquation = depending.getEquation().getEquation();
            boolean sum = false;
            if (depEquation[0].startsWith("SUM"))
                sum = true;
            else
                continue;
            // count stays the same, squares don't get shot even if birds do :-)
            
            if (variable.getEntity().isMatrix() == false &&
                    depending.getEntity().isMatrix() == false) {
                // one-dim affecting an one-dim spatial, possible
                
                
                result.add("! 1 to number of neighbours");
                result.add("DO k = 1, spatial(i, 1)");
                if (variable.isInteger())
                    result.add("sum_int = 0");
                else
                    result.add("sum_real = 0");
                result.add("neighbour = spatial(i, k + 1)");
                result.add("! updating " + depname + " % one_dim(spatial(i, k + 1))");
                result.add("DO h = 1, spatial(neighbour, 1)"); // neighbours of neighbour
                if (variable.isInteger())
                    result.add("sum_int = sum_int + "+ name +" % one_dim(spatial(neighbour, h + 1)) % value");
                else
                    result.add("sum_real = sum_real + "+ name +" % one_dim(spatial(neighbour, h + 1)) % value");
                
                result.add("END DO");
                if (variable.isInteger())
                    result.add(depname + " % one_dim(neighbour) % value = sum_int");
                else
                    result.add(depname + " % one_dim(neighbour) % value = sum_real");
                result.add("END DO");
                
            } else if (variable.getEntity().isMatrix() == false &&
                    depending.getEntity().isMatrix() == true) {
                // one-dim affecting a two-dim spatial, impossible?
                throw new InvalidModelException("Acceptation: A one-dim variable is affecting a two-dim spatial!");
                
            } else {
                
                // both 2-dimensional, possible (the same entity)
                // two dimensional sum case...
                // when variable and dep have the same number of dimensions
                
                String correct_sum = "";
                if (variable.isInteger()) {
                    correct_sum = "sum_int";
                } else correct_sum = "sum_real";
                if (variable.getEntity().getYCoordinate().isSpatial()){
                    result.add("! 1 to number of neighbours");
                    result.add("DO k = 1, spatial(i, 1)");
                    result.add(correct_sum + " = 0");
                    result.add("neighbour = spatial(i, k + 1)");
                    result.add("! updating " + depname + " % two_dim(spatial(i, k + 1), j)");
                    result.add("DO h = 1, spatial(neighbour, 1)"); // neighbours of neighbour
                    result.add(correct_sum + " = " + correct_sum + " + " + name + " % two_dim(spatial(neighbour, h + 1), j) % value");
                    result.add("END DO");
                    result.add(depname + " % two_dim(neighbour, j) % value = " + correct_sum);
                    result.add("END DO");
                } else{
                    result.add("! 1 to number of neighbours");
                    result.add("DO k = 1, spatial(j, 1)");
                    result.add(correct_sum + " = 0");
                    result.add("neighbour = spatial(j, k + 1)");
                    result.add("! updating " + depname + " % two_dim(i, spatial(j, k + 1))");
                    result.add("DO h = 1, spatial(neighbour, 1)"); // neighbours of neighbour
                    result.add(correct_sum + " = " + correct_sum + " + " + name + " % two_dim(i, spatial(neighbour, h + 1)) % value");
                    result.add("END DO");
                    result.add(depname + " % two_dim(i, neighbour) % value = " + correct_sum);
                    result.add("END DO");
                    
                }
            }
        }
        
        
        
        
        result.add("END IF");
        result.add("prob_index = prob_index + 1");
        result.add("IF (prob_index > SIZE(prob) .OR. prob_index < 1) THEN");
        result.add("CALL G05FAF(0.0_dp, 1.0_dp, SIZE(prob), prob)");
        result.add("prob_index = 1");
        result.add("END IF");
        
        return result;
    }
    
    /**
     * Sets correct indexing for functional variables with
     * spatial dependencies. Used for updating neighbouring square functional properties after
     * the spatial functional variable target has been updated.
     * For example: q = sum(&x) and p = q/2.0.
     * When x is updated (the spatial target)
     * first q is calculated for all neighbours of x,
     * then all neighbours of x get a new p to match their new q values.
     */
    
    private static ArrayList<String> generateSpatialDeps(Variable functional, String indexing, boolean spatialY){
        ArrayList<String> result = new ArrayList<String>();
        String[] equation = functional.getEquation().getEquation();
        
        
        Variable[] parameters = functional.getEquation().getParameters();
        
        
        // DEBUG
        /*
        System.out.println("GENERATOR DEBUG: generateSpatialDeps for " + functional.getName());
        System.out.print("newToBeUsed contains:");
        for (Variable n : newToBeUsed){
            System.out.print(n.getName() + " ");
        }
        System.out.println();
         */
        
        // check the equation of depending
        for (int i=0; i<parameters.length; i++) {
            // dealing with Variable parameters[i]
            
            //System.out.println("GENERATOR DEBUG: dealing with " + parameters[i].getName());
            if (parameters[i].getEntity() == null) { // global
                for (int j=0; j<equation.length; j++) {
                    if (equation[j].equals(parameters[i].getName())) {
                        equation[j] += " % one_dim(1) % value";
                    }
                } // end looping the equation
            } else if (parameters[i].getEntity().isMatrix() == false) { // one-dimensional
                if (parameters[i].getEntity().equals(functional.getEntity().getXCoordinate())) {
                    for (int j=0; j<equation.length; j++) {
                        if (equation[j].equals(parameters[i].getName())) {
                            if (spatialY)
                                equation[j] += " % one_dim(j) % value";
                            else
                                equation[j] += " % one_dim(neighbour) % value";
                        }
                    } // end looping the equation
                } else { // either the same entity or equals y-coordinate
                    for (int j=0; j<equation.length; j++) {
                        if (equation[j].equals(parameters[i].getName())) {
                            if (spatialY)
                                equation[j] += " % one_dim(neighbour) % value";
                            else
                                equation[j] += " % one_dim(i) % value";
                        }
                    } // end looping the equation
                }
                
            } else { // two-dimensional
                for (int j=0; j<equation.length; j++) {
                    if (equation[j].equals(parameters[i].getName())) {
                        if (spatialY)
                            equation[j] += " % two_dim(neighbour, j) % value";
                        else
                            equation[j] += " % two_dim(i, neighbour) % value";
                    }
                } // end looping the equation
            }
        } // end looping the parameters[]
        
        String finalEquation = "";
        for (int k=0; k<equation.length; k++) {
            finalEquation = finalEquation + equation[k];
        }
        
        result.add(functional.getName() + indexing + " % value = " + finalEquation);
        
        return result;
        
    }
    
    private static ArrayList<String> acceptationAffectedOf(Variable variable, Variable depending){
        ArrayList<String> result = new ArrayList<String>();
        if (variable.getEntity() == null && depending.getEntity() == null) {
            result.addAll(generateUpdateOneFunctional(depending));
        } else if (variable.getEntity() == null &&
                depending.getEntity().isMatrix() == false) {
            result.add("DO i2 = 1, " + depending.getEntity().getSize());
            result.addAll(generateUpdateOneFunctional(depending));
            result.add("END DO");
        } else if (variable.getEntity() == null &&
                depending.getEntity().isMatrix()) {
            result.add("DO i2 = 1, " + depending.getEntity().getYCoordinate().getSize());
            result.add("DO j2 = 1, " + depending.getEntity().getXCoordinate().getSize());
            result.addAll(generateUpdateOneFunctional(depending));
            result.add("END DO");
            result.add("END DO");
        } else if (variable.getEntity().isMatrix() == false &&
                depending.getEntity().isMatrix() == false) {
            result.add("i2 = i");
            result.addAll(generateUpdateOneFunctional(depending));
        } else if (variable.getEntity().isMatrix() == false &&
                depending.getEntity().isMatrix() == true) {
            if (variable.getEntity().equals(depending.getEntity().getYCoordinate())) {
                result.add("i2 = i");
                result.add("DO j2 = 1, " +
                        depending.getEntity().getXCoordinate().getSize());
                result.addAll(generateUpdateOneFunctional(depending));
                result.add("END DO");
            } else {
                result.add("j2 = i");
                result.add("DO i2 = 1, " +
                        depending.getEntity().getYCoordinate().getSize());
                result.addAll(generateUpdateOneFunctional(depending));
                result.add("END DO");
            }
        } else {
            result.add("i2 = i");
            result.add("j2 = j");
            result.addAll(generateUpdateOneFunctional(depending));
        }
        return result;
    }
    
    private static ArrayList<String> generateUpdateOneFunctional(Variable depending) {
        
        // This generates the code that replaces the value by new_value for one particular variable
        
        // depending is indexed: (1), (i) or (i2, j2).
        
        assert depending.isFunctional() == true;
        
        ArrayList<String> result=new ArrayList<String>();
        String indexing = "";
        
        if (depending.getEntity() == null) {
            indexing = "one_dim(1)";
        } else if (depending.getEntity().isMatrix() == false){
            indexing = "one_dim(i2)";
        } else indexing = "two_dim(i2, j2)";
        
        result.add(depending.getName() + " % " + indexing + " % value = " +
                depending.getName() + " % " + indexing + " % new_value");
        
        return result;
        
    }
    
    
    private static ArrayList<Variable> topologicalSort(ArrayList<Variable> sortThis) {
        // This method sorts the list sortThis topologically.
        // We assume that the variables are introduced in topological order in the model specification.
        assert topologicalList != null;
        
        ArrayList<Variable> result = new ArrayList<Variable>(topologicalList);
        result.retainAll(sortThis);
        return result;
    }
    
    private static ArrayList<String> generateNewValueForOneFunctional(Variable functional,
            Set<Variable> newToBeUsed) throws InvalidModelException {
        // this is a different approach to generateNewValuesFunctionalCode
        
        // We assume that functional is indexed by one_dim(1), one_dim(i2) or two_dim(i2, j2)
        // and generate the code that calculates the new value and places it to new_value.
        
        // Note that the spatial functional parameters don't need new values here, so this method is not called.
        
        ArrayList<String> result = new ArrayList<String>();
        
        //String[] equation = functional.getEquation().getEquation(); A BUG!
        
        String[] equation = functional.getEquation().getEquation();
        
        /* the equation in parts: for example:
        3 * alpha + EXP(beta+gamma) ->
        3 *
        alpha
        + EXP(
        beta
        +
        gamma
        )
         */
        
        Variable[] parameters = functional.getEquation().getParameters();
        
        
        // DEBUG
        /*
        System.out.println("GENERATOR DEBUG: generateNewValueForOneFunctional for " + functional.getName());
        System.out.print("newToBeUsed contains:");
        for (Variable n : newToBeUsed){
            System.out.print(n.getName() + " ");
        }
        System.out.println();
         */
        
        //result.add("! DEBUG: a mysterious bug resides here somewhere.");
        // check the equation of depending
        for (int i=0; i<parameters.length; i++) {
            // dealing with Variable parameters[i]
            
            //System.out.println("GENERATOR DEBUG: dealing with " + parameters[i].getName());
            if (parameters[i].getEntity() == null) { // global
                for (int j=0; j<equation.length; j++) {
                    if (equation[j].equals(parameters[i].getName())) {
                        if (newToBeUsed.contains(parameters[i])) {
                            //System.out.println("here 1");
                            equation[j] += " % one_dim(1) % new_value";
                        } else {
                            //System.out.println("here 2");
                            equation[j] += " % one_dim(1) % value";
                        }
                    }
                } // end looping the equation
            } else if (parameters[i].getEntity().isMatrix() == false) { // one-dimensional
                if (parameters[i].getEntity().equals(functional.getEntity().getXCoordinate())) {
                    for (int j=0; j<equation.length; j++) {
                        if (equation[j].equals(parameters[i].getName())) {
                            if (newToBeUsed.contains(parameters[i])) {
                                //System.out.println("here 3");
                                equation[j] += " % one_dim(j2) % new_value";
                            } else {
                                //System.out.println("here 4");
                                equation[j] += " % one_dim(j2) % value";
                            }
                        }
                    } // end looping the equation
                } else { // either the same entity or equals y-coordinate
                    for (int j=0; j<equation.length; j++) {
                        if (equation[j].equals(parameters[i].getName())) {
                            if (newToBeUsed.contains(parameters[i])) {
                                //System.out.println("here 5");
                                equation[j] += " % one_dim(i2) % new_value";
                            } else {
                                //System.out.println("here 6");
                                equation[j] += " % one_dim(i2) % value";
                            }
                        }
                    } // end looping the equation
                }
                
            } else { // two-dimensional
                for (int j=0; j<equation.length; j++) {
                    if (equation[j].equals(parameters[i].getName())) {
                        if (newToBeUsed.contains(parameters[i])) {
                            //System.out.println("here 7");
                            equation[j] += " % two_dim(i2, j2) % new_value";
                        } else {
                            //System.out.println("here 8");
                            equation[j] += " % two_dim(i2, j2) % value";
                        }
                    }
                } // end looping the equation
            }
        } // end looping the parameters[]
        
        String finalEquation = "";
        for (int k=0; k<equation.length; k++) {
            finalEquation = finalEquation + equation[k];
        }
        
        if (functional.getEntity() == null) {
            result.add(functional.getName() + " % one_dim(1) % new_value = " + finalEquation);
        } else if (functional.getEntity().isMatrix() == false) {
            result.add(functional.getName() + " % one_dim(i2) % new_value = " + finalEquation);
        } else result.add(functional.getName() + " % two_dim(i2, j2) % new_value = " + finalEquation);
        
        return result;
    }
    
    /**
     * Generates code that calculates new values for all the functional parameters depending on
     * the current parameter.
     *
     * @param  variable The current parameter. We assume that it
     * is indexed by "one_dim(1)" if it is global, by "one_dim(i)" if it is one-dimensional and by
     * "two_dim(i, j)" if it is two-dimensional. Note that the new values for "SUM"-parameters are
     * not calculated here.
     */
    public static ArrayList<String> generateNewValuesFunctionalCode(Variable variable)
    throws InvalidModelException {
        
        ArrayList<String> result = new ArrayList<String>();
        ArrayList<Variable> allAffected = new ArrayList<Variable>(variable.getAffectsList());
        Set<Variable> newToBeUsed = new HashSet<Variable>();
        newToBeUsed.add(variable);
        
        //System.out.println("GENERATOR DEBUG: generateNewValuesFunctionalCode for " + variable.getName());
        
        for (int iterator = 0; iterator < allAffected.size(); iterator++) {
            Variable depending = allAffected.get(iterator);
            
            //System.out.println("GENERATOR DEBUG: depending (phase 1) is : " + depending.getName());
            // not functional
            if (depending.isFunctional() == false) continue;
            
            // the spatial parameters won't need new values here and no new dependencies are added
            if (depending.isSpatial() == true) continue;
            
            // add further affected variables to all_depending
            allAffected.addAll(depending.getAffectsList());
            newToBeUsed.add(depending);
        }
        
        // DEBUG
        /*
        if (allAffected == null) {
            System.out.println("GENERATOR DEBUG: allAffected is null.");
        }*/
        
        allAffected = topologicalSort(allAffected);
        
        for (int iterator = 0; iterator < allAffected.size(); iterator++) {
            
            Variable depending = allAffected.get(iterator);
            
            //System.out.println("GENERATOR DEBUG: depending (phase 2) is : " + depending.getName());
            
            // not functional
            if (depending.isFunctional() == false) continue;
            if (depending.isSpatial()) continue; // we don't need to calculate new values for them
            
            result.add("");
            result.add("! calculating new value(s) for " + depending.getName());
            result.add("");
            
            if (variable.getEntity() == null && depending.getEntity() == null) {
                // no loops
                result.addAll(generateNewValueForOneFunctional(depending, newToBeUsed));
                return result;
            }
            
            if (variable.getEntity() == null && depending.getEntity().isMatrix() == false){
                // one loop
                result.add("DO i2 = 1, " + depending.getEntity().getSize());
                result.addAll(generateNewValueForOneFunctional(depending, newToBeUsed));
                result.add("END DO");
                return result;
            }
            
            if (variable.getEntity() == null && depending.getEntity().isMatrix() == true) {
                // two loops
                result.add("DO i2 = 1, " + depending.getEntity().getYCoordinate().getSize());
                result.add("DO j2 = 1, " + depending.getEntity().getXCoordinate().getSize());
                result.addAll(generateNewValueForOneFunctional(depending, newToBeUsed));
                result.add("END DO");
                result.add("END DO");
                return result;
            }
            
            if (variable.getEntity().isMatrix() == false && depending.getEntity().isMatrix() == false) {
                // no loops
                result.add("i2 = i");
                result.addAll(generateNewValueForOneFunctional(depending, newToBeUsed));
                return result;
            }
            
            if (variable.getEntity().isMatrix() == false && depending.getEntity().isMatrix() == true) {
                // one loop
                if (variable.getEntity().equals(depending.getEntity().getYCoordinate())) {
                    result.add("i2 = i");
                    result.add("DO j2 = 1, " + depending.getEntity().getXCoordinate().getSize());
                    result.addAll(generateNewValueForOneFunctional(depending, newToBeUsed));
                    result.add("END DO");
                    return result;
                } else {
                    result.add("j2 = i");
                    result.add("DO i2 = 1, " + depending.getEntity().getYCoordinate().getSize());
                    result.addAll(generateNewValueForOneFunctional(depending, newToBeUsed));
                    result.add("END DO");
                    return result;
                }
            }
            
            if (variable.getEntity().isMatrix() == true && depending.getEntity().isMatrix() == true) {
                // no loops
                result.add("i2 = i");
                result.add("j2 = j");
                result.addAll(generateNewValueForOneFunctional(depending, newToBeUsed));
                return result;
            }
        }
        
        return result;
    }
    
    
    /**
     * Generates code that calculates the acceptance probability for the current parameter.
     *
     * @param  variable The current parameter. We assume that it
     * is indexed by "one_dim(1)" if it is global, by "one_dim(i)" if it is one-dimensional and by
     * "two_dim(i, j)" if it is two-dimensional.
     */
    public static ArrayList<String> generateAcceptationFormula(Variable variable)
    throws InvalidModelException, IllegalParametersException, MissingFunctionException {
        
        /* The idea:
         
        1. probability for this parameter
        2. probabilities for all affected variables
        3. q-probabilities.
         
        Note: This method DOES NOT generate:
        - the loops (acceptation for all x's for example)
        - generating proposals
        - acceptation
        - updating functional parameters
        or anything like that.
         */
        
        assert variable.isFunctional() == false;
        assert variable.isData() == false || variable.getMissingValueCount() > 0;
        
        if (variable.getEntity() == null) {
            return generateAcceptationFormulaGlobal(variable);
        } else if (variable.getEntity().isMatrix() == false) {
            return generateAcceptationFormulaOneDimensional(variable);
        } else {
            return generateAcceptationFormulaTwoDimensional(variable);
        }
        
    }
    
    private static ArrayList<String> generateAcceptationFormulaGlobal(Variable variable)
    throws InvalidModelException, IllegalParametersException, MissingFunctionException {
        
        //System.out.println("GENERATOR DEBUG: generateAcceptationFormulaGlobal");
        /*
          This variable is global.
          It is indexed like variable % one_dim(1).
         
          All affecting variables must also be global.
         */
        
        ArrayList<String> result = new ArrayList<String>();
        
        int parameterCount = variable.getDistribution().getNumberOfParameters();
        
        String[] indices = new String[parameterCount+2];
        
        // 1. P(variable') / P(variable)
        
        // DEBUG
        //System.out.println("GENERATOR DEBUG: 1. P(" + variable.getName() + "') / P(" + variable.getName() + ")");
        
        result.add("");
        result.add("!  P(" + variable.getName() + "') / P(" + variable.getName() + ")");
        result.add("");
        
        result.add("p_acc = 1");
        
        for (int i=0; i<parameterCount; i++) {
            Object parameter = variable.getDistribution().getParameter(i);
            
            if (parameter instanceof Integer) {
                indices[i] = ((Integer)parameter).toString();
                continue;
            } else if (parameter instanceof Double) {
                // we must add _dp to the real constants
                indices[i] = ((Double)parameter).toString()+"_dp";
                continue;
            }
            // now we know that the parameter is a Variable.
            
            Variable varParameter=(Variable)parameter;
            // it must be global
            indices[i] = varParameter.getName()+" % one_dim(1) % value";
        }
        indices[parameterCount] = variable.getName() + " % one_dim(1) % new_value";
        indices[parameterCount+1] = "new_frequency";
        
        // P(variable')
        
        // DEBUG
        /*
          System.out.println("GENERATOR DEBUG: About to call getFreqCode with parameters: ");
          for (int print=0; print<indices.length; print++)
          {
          System.out.println(indices[print]);
          }
         */
        
        ArrayList<String> toAdd=variable.getDistribution().getFreqCode(indices);
        result.addAll(toAdd);
        
        indices[parameterCount] = variable.getName() + " % one_dim(1) % value";
        indices[parameterCount+1] = "frequency";
        
        // P(variable)
        
        // DEBUG
        /*
          System.out.println("GENERATOR DEBUG: About to call getFreqCode with parameters: ");
          for (int print=0; print<indices.length; print++)
          {
          System.out.println(indices[print]);
          }
         */
        
        toAdd=variable.getDistribution().getFreqCode(indices);
        result.addAll(toAdd);
        
        result.add("p_acc = p_acc * new_frequency / frequency");
        
        // 2) P(x | variable') / P(x | variable) for all x that depend on variable
        
        // DEBUG
        /*
          System.out.println("GENERATOR DEBUG: Phase 2. P( depending | " + variable.getName() +
          ") / P( depending | " + variable.getName() + "')");
         */
        
        ArrayList<Variable> allAffected = new ArrayList<Variable>(variable.getAffectsList());
        Set<Variable> newToBeUsed = new HashSet<Variable>();
        newToBeUsed.add(variable);
        
        Set<Variable> handledAlready = new HashSet<Variable>();

        for (int iterator = 0; iterator < allAffected.size(); iterator++) {
            Variable depending = allAffected.get(iterator);
            
            if (depending.isSpatial()) continue;
            
            // functional but not spatial
            if (depending.isFunctional()) {
                // add further affected variables to all_depending
                allAffected.addAll(depending.getAffectsList());
                newToBeUsed.add(depending);
                continue;
            }

            if (depending.equals(variable)) continue; // no loops please
            
            // not functional
            
            // DEBUG
            /*
            System.out.println("GENERATOR DEBUG: Phase 2, in detail: P( " + depending.getName() + " | "
                + variable.getName() + "') / P( " + depending.getName() + " | " + variable.getName() + ")");
             */
            
            if (handledAlready.contains(depending)) continue; // no duplicates

            result.add("");
            result.add("! P( " + depending.getName() + " | "
                    + variable.getName() + "') / P( " + depending.getName() + " | " + variable.getName() + ")");
            result.add("");

            handledAlready.add(depending);
            
            // now variable is global, so depending may be global, one-dimensional or two-dimensional
            
            if (depending.getEntity() == null) {
                // depending is global
                result.addAll(generateLikelihoodFormulaGlobal(depending, newToBeUsed));
                
            } else if (depending.getEntity().isMatrix() == false) {
                // depending is one-dimensional
                
                // we have to generate a loop
                result.add("DO i=1, "+depending.getEntity().getSize());
                result.addAll(generateLikelihoodFormulaOneDimensional(depending, newToBeUsed));
                result.add("END DO");
            }
            
            else {
                // depending is two-dimensional
                
                // we have to generate two loops
                result.add("DO i=1, "+depending.getEntity().getYCoordinate().getSize());
                result.add("DO j=1, "+depending.getEntity().getXCoordinate().getSize());
                result.addAll(generateLikelihoodFormulaTwoDimensional(depending, newToBeUsed, true));
                result.add("END DO");
                result.add("END DO");
            }
        }
        
        result.addAll(generateTransitionFormula(variable));
        return result;
    }
    
    private static ArrayList<String> generateAcceptationFormulaOneDimensional(Variable variable)
    throws InvalidModelException, IllegalParametersException, MissingFunctionException {
        
        //System.out.println("GENERATOR DEBUG: generateAcceptationFormulaOneDimensional");
        /*
          This variable belongs to an one-dimensional entity.
          It is indexed like variable % one_dim(i).
         
          All affecting variables must be either global
          or belong to the same one-dimensional entity.
         
          All affected variables must belong to the same one-dimensional
          entity or to the two-dimensional child entity of variable's entity.
         */
        
        // DEBUG
        /*
          System.out.println("GENERATOR DEBUG: called generateAcceptationFormulaOneDimensional with " +
          variable.getName());
         */
        
        ArrayList<String> result = new ArrayList<String>();
        
        int parameterCount = variable.getDistribution().getNumberOfParameters();
        
        String[] indices = new String[parameterCount+2];
        
        // DEBUG
        //System.out.println("GENERATOR DEBUG: 1. P(" + variable.getName() + ") / P(" + variable.getName() + "')");
        
        result.add("");
        result.add("!  P(" + variable.getName() + "') / P(" + variable.getName() + ")");
        result.add("");
        
        result.add("p_acc = 1");
        
        for (int i=0; i<parameterCount; i++) {
            Object parameter = variable.getDistribution().getParameter(i);
            
            if (parameter instanceof Integer) {
                indices[i] = ((Integer)parameter).toString();
                continue;
            } else if (parameter instanceof Double) {
                // we must add _dp to the real constants
                indices[i] = ((Double)parameter).toString()+"_dp";
                continue;
            }
            // now we know that the parameter is a Variable.
            
            Variable varParameter=(Variable)parameter;
            
            if (varParameter.getEntity() == null) {
                indices[i] = varParameter.getName() + " % one_dim(1) % value";
            } else if (varParameter.getEntity().equals(variable.getEntity())) {
                indices[i] = varParameter.getName() + " % one_dim(i) % value";
            } else throw new InvalidModelException("Variable " +
                    varParameter.getName() + " affects variable " + variable.getName());
        }
        indices[parameterCount] = variable.getName()+" % one_dim(i) % new_value";
        indices[parameterCount+1] = "new_frequency";
        
        // P(variable')
        
        // DEBUG
        /*
          System.out.println("GENERATOR DEBUG: About to call getFreqCode with parameters: ");
          for (int print=0; print<indices.length; print++)
          {
          System.out.println(indices[print]);
          }
         */
        
        ArrayList<String> toAdd = variable.getDistribution().getFreqCode(indices);
        result.addAll(toAdd);
        
        indices[parameterCount] = variable.getName()+" % one_dim(i) % value";
        indices[parameterCount+1] = "frequency";
        
        // P(variable)
        
        // DEBUG
        /*
          System.out.println("GENERATOR DEBUG: About to call getFreqCode with parameters: ");
          for (int print=0; print<indices.length; print++)
          {
          System.out.println(indices[print]);
          }
         */
        
        toAdd = variable.getDistribution().getFreqCode(indices);
        result.addAll(toAdd);
        
        result.add("p_acc = p_acc * new_frequency / frequency");
        
        // 2) P(x | variable') / P(x | variable) for all x that depend on variable
        
        // DEBUG
        /*
          System.out.println("GENERATOR DEBUG: Phase 2. P( depending | " + variable.getName() +
          ") / P( depending | " + variable.getName() + "')");
         */
        
        result.add("");
        
        ArrayList<Variable> allAffected = new ArrayList<Variable>(variable.getAffectsList());
        Set<Variable> newToBeUsed = new HashSet<Variable>();
        newToBeUsed.add(variable);
        
        Set<Variable> handledAlready = new HashSet<Variable>();

        for (int iterator = 0; iterator < allAffected.size(); iterator++) {
            Variable depending = allAffected.get(iterator);
            
            if (depending.isSpatial()) continue;
            
            // functional
            if (depending.isFunctional()) {
                
                // add further affeced variables to all_depending
                allAffected.addAll(depending.getAffectsList());
                newToBeUsed.add(depending);
                continue;
            }

            if (depending.equals(variable)) continue; // no loops please

            if (handledAlready.contains(depending)) continue; // no duplicates
            
            // not functional
            
            // DEBUG
            /*
            System.out.println("GENERATOR DEBUG: Phase 2, in detail: P( " + depending.getName() + " | "
                + variable.getName() + "') / P( " + depending.getName() + " | " + variable.getName() + ")");
             */
            
            result.add("");
            result.add("! P( " + depending.getName() + " | "
                    + variable.getName() + "') / P( " + depending.getName() + " | " + variable.getName() + ")");
            result.add("");
            
            // now variable is one-dimensional, so depending may be one-dimensional or two-dimensional
            
            handledAlready.add(depending);

            if (depending.getEntity() == null) {
                throw new InvalidModelException("One-dimensional variable " +
                        variable.getName() + " affects global variable " + depending.getName());
            } else if (depending.getEntity().isMatrix() == false) {
                // depending is one-dimensional
                
                // depending's entity must be the same as variable's entity.
                if (!depending.getEntity().equals(variable.getEntity())) {
                    throw new InvalidModelException("Variable " +
                            variable.getName() + " affects variable " + depending.getName());
                }
                
                result.addAll(generateLikelihoodFormulaOneDimensional(depending, newToBeUsed));
            } else {
                // depending is two-dimensional
                
                // we have to generate a loop
                if (depending.getEntity().getXCoordinate().equals(variable.getEntity())) {
                    result.add("DO j=1, "+depending.getEntity().getYCoordinate().getSize());
                    
                    // depending is indexed depending % two_dim(j, i)
                    result.addAll(generateLikelihoodFormulaTwoDimensional(depending, newToBeUsed, false));
                    result.add("END DO");
                    
                } else if (depending.getEntity().getYCoordinate().equals(variable.getEntity())) {
                    result.add("DO j=1, "+depending.getEntity().getXCoordinate().getSize());
                    
                    // depending is indexed depending % two_dim(i, j)
                    result.addAll(generateLikelihoodFormulaTwoDimensional(depending, newToBeUsed, true));
                    result.add("END DO");
                } else {
                    throw new InvalidModelException("Variable " +
                            variable.getName() + " affects variable " + depending.getName());
                }
            }
        }
        result.addAll(generateTransitionFormula(variable));
        return result;
    }
    
    private static ArrayList<String> generateAcceptationFormulaTwoDimensional(Variable variable)
    throws InvalidModelException, IllegalParametersException, MissingFunctionException {
        
        //System.out.println("GENERATOR DEBUG: generateAcceptationFormulaTwoDimensional");
        /*
          This variable belongs to a two-dimensional entity.
          It is indexed like variable % two_dim(i, j).
         
          All affecting variables may be global,
          belong to the vertical entity, belong to the horizontal entity,
          or belong to the same two-dimensional entity.
         
          All affected variables must belong to the same entity.
         */
        
        ArrayList<String> result = new ArrayList<String>();
        
        int parameterCount = variable.getDistribution().getNumberOfParameters();
        
        String[] indices = new String[parameterCount+2];
        
        // 1. P(variable') / P(variable)
        
        // DEBUG
        //System.out.println("GENERATOR DEBUG: 1. P(" + variable.getName() + ") / P(" + variable.getName() + "')");
        
        result.add("");
        result.add("!  P(" + variable.getName() + "') / P(" + variable.getName() + ")");
        result.add("");
        
        result.add("p_acc = 1");
        
        for (int i=0; i<parameterCount; i++) {
            //int type = variable.getDistribution().getParameter(i);
            Object parameter = variable.getDistribution().getParameter(i);
            
            if (parameter instanceof Integer) {
                indices[i] = ((Integer)parameter).toString();
                continue;
            } else if (parameter instanceof Double) {
                // we must add _dp to the real constants
                indices[i] = ((Double)parameter).toString()+"_dp";
                continue;
            }
            // now we know that the parameter is a Variable.
            
            Variable varParameter=(Variable)parameter;
            
            if (varParameter.getEntity() == null) {
                indices[i] = varParameter.getName()+" % one_dim(1) % value";
            } else if (varParameter.getEntity().equals(variable.getEntity().getYCoordinate())) {
                indices[i] = varParameter.getName()+" % one_dim(i) % value";
            } else if (varParameter.getEntity().equals(variable.getEntity().getXCoordinate())) {
                indices[i] = varParameter.getName()+" % one_dim(j) % value";
            } else if (varParameter.getEntity().equals(variable.getEntity())) {
                indices[i] = varParameter.getName()+" % two_dim(i, j) % value";
            }
        }
        indices[parameterCount] = variable.getName()+" % two_dim(i, j) % new_value";
        indices[parameterCount+1] = "new_frequency";
        
        // P(variable')
        
        // DEBUG
        /*
          System.out.println("GENERATOR DEBUG: About to call getFreqCode with parameters: ");
          for (int print=0; print<indices.length; print++)
          {
          System.out.println(indices[print]);
          }
         */
        
        ArrayList<String> toAdd=variable.getDistribution().getFreqCode(indices);
        result.addAll(toAdd);
        
        indices[parameterCount] = variable.getName()+" % two_dim(i, j) % value";
        indices[parameterCount+1] = "frequency";
        
        // P(variable)
        
        // DEBUG
        /*
          System.out.println("GENERATOR DEBUG: About to call getFreqCode with parameters: ");
          for (int print=0; print<indices.length; print++)
          {
          System.out.println(indices[print]);
          }
         */
        
        toAdd=variable.getDistribution().getFreqCode(indices);
        result.addAll(toAdd);
        
        result.add("p_acc = p_acc * new_frequency / frequency");
        
        // 2) P(x | variable') / P(x | variable) for all x that depend on variable
        
        // DEBUG
        /*
          System.out.println("GENERATOR DEBUG: Phase 2. P( depending | " + variable.getName() +
          ") / P( depending | " + variable.getName() + "')");
         */
        
        ArrayList<Variable> allAffected = new ArrayList<Variable>(variable.getAffectsList());
        Set<Variable> newToBeUsed = new HashSet<Variable>();
        newToBeUsed.add(variable);

        Set<Variable> handledAlready = new HashSet<Variable>();
        
        for (int iterator = 0; iterator < allAffected.size(); iterator++) {
            Variable depending = allAffected.get(iterator);
            
            if (depending.isSpatial()) continue;

            // functional
            if (depending.isFunctional()) {
                // add further affeced variables to all_depending
                allAffected.addAll(depending.getAffectsList());
                newToBeUsed.add(depending);
                continue;
            }
        
            
            if (depending.equals(variable)) continue; // no loops please

            if (handledAlready.contains(depending)) continue; // no duplicates

            // not functional
            
            // DEBUG
            /*
                System.out.println("GENERATOR DEBUG: Phase 2, in detail: P( " + depending.getName() + " | "
                    + variable.getName() + ") / P( " + depending.getName() + " | " + variable.getName() + "')");
             */
            
            result.add("");
            result.add("! P( " + depending.getName() + " | "
                    + variable.getName() + "') / P( " + depending.getName() + " | " + variable.getName() + ")");
            result.add("");

            handledAlready.add(depending);
            
            // now variable is two-dimensional, so depending must be two-dimensional
            
            if (depending.getEntity() == null) {
                throw new InvalidModelException("Two-dimensional variable " +
                        variable.getName() + " affects global variable " + depending.getName());
            } else if (depending.getEntity().isMatrix() == false) {
                throw new InvalidModelException("Two-dimensional variable " +
                        variable.getName() + " affects one-dimensional variable " + depending.getName());
            } else if (depending.getEntity().equals(variable.getEntity()) == false) {
                throw new InvalidModelException("Variable " +
                        variable.getName() + " affects " + depending.getName() + " belonging to a different entity");
            } else {
                // depending is two-dimensional and its entity is the same as variable's
                
                result.addAll(generateLikelihoodFormulaTwoDimensional(depending, newToBeUsed, true));
            }
        }
        
        result.addAll(generateTransitionFormula(variable));
        
        return result;
    }
    
    private static ArrayList<String> generateLikelihoodFormulaGlobal(Variable depending, Set<Variable> newToBeUsed)
    throws IllegalParametersException, InvalidModelException, MissingFunctionException {
        
        //System.out.println("GENERATOR DEBUG: generateLikelihoodFormulaGlobal");
        // This method returns code that calculates the likelihood of depending, when using new values of variables in newToBeUsed
        ArrayList<String> result = new ArrayList<String>();
        String[] indices;
        int parameterCount = depending.getDistribution().getNumberOfParameters();
        indices = new String[parameterCount+2];
        
        for (int i=0; i<parameterCount; i++) {
            Object parameter = depending.getDistribution().getParameter(i);
            
            if (parameter instanceof Integer) {
                indices[i] = ((Integer)parameter).toString();
                continue;
            } else if (parameter instanceof Double) {
                // we must add _dp to the real constants
                indices[i] = ((Double)parameter).toString()+"_dp";
                continue;
            }
            // now we know that the parameter is a Variable.
            
            Variable varParameter=(Variable)parameter;
            
            if (varParameter.getEntity() != null) {
                throw new InvalidModelException("Variable " +
                        varParameter.getName() + " affects variable " + depending.getName());
            }
            
            if (newToBeUsed.contains(varParameter)) {
                indices[i] = varParameter.getName() + "% one_dim(1) % new_value";
            } else {
                indices[i] = varParameter.getName()+" % one_dim(1) % value";
            }
        }
        indices[parameterCount] = depending.getName()+" % one_dim(1)";
        indices[parameterCount+1] = "new_frequency";
        
        // P(depending | variable')
        
        // DEBUG
        /*
          System.out.println("GENERATOR DEBUG: About to call getFreqCode with parameters: ");
          for (int print=0; print<indices.length; print++) {
          System.out.println(indices[print]);
          }
         */
        
        ArrayList<String> toAdd=depending.getDistribution().getFreqCode(indices);
        result.addAll(toAdd);
        
        for (int i=0; i<parameterCount; i++) {
            indices[i] = indices[i].replace("% new_value", "% value");
        }
        
        indices[parameterCount+1] = "frequency";
        
        // P(depending | variable)
        
        // DEBUG
        /*
          System.out.println("GENERATOR DEBUG: About to call getFreqCode with parameters: ");
          for (int print=0; print<indices.length; print++) {
          System.out.println(indices[print]);
          }
         */
        
        toAdd=depending.getDistribution().getFreqCode(indices);
        result.addAll(toAdd);
        
        result.add("p_acc = p_acc * new_frequency / frequency");
        
        return result;
    }
    
    private static ArrayList<String> generateLikelihoodFormulaOneDimensional(Variable depending,
            Set<Variable> newToBeUsed) throws IllegalParametersException, InvalidModelException,
            MissingFunctionException {
        
        //System.out.println("GENERATOR DEBUG: generateLikelihoodFormulaOneDimensional");
        ArrayList<String> result = new ArrayList<String>();
        String[] indices;
        int parameterCount = depending.getDistribution().getNumberOfParameters();
        indices = new String[parameterCount+2];
        
        for (int i=0; i<parameterCount; i++) {
            Object parameter = depending.getDistribution().getParameter(i);
            
            if (parameter instanceof Integer) {
                indices[i] = ((Integer)parameter).toString();
                continue;
            } else if (parameter instanceof Double) {
                // we must add _dp to the real constants
                indices[i] = ((Double)parameter).toString()+"_dp";
                continue;
            }
            // now we know that the parameter is a Variable.
            
            Variable varParameter=(Variable)parameter;
            
            boolean useNew = false;
            
            if (newToBeUsed.contains(varParameter)) useNew = true;
            
            if (varParameter.getEntity() == null) {
                if (useNew)
                    indices[i] = varParameter.getName() + " % one_dim(1) % new_value";
                else
                    indices[i] = varParameter.getName() + " % one_dim(1) % value";
            } else if (varParameter.getEntity().equals(depending.getEntity())) {
                if (useNew)
                    indices[i] = varParameter.getName() + " % one_dim(i) % new_value";
                else
                    indices[i] = varParameter.getName() + " % one_dim(i) % value";
            }
        }
        indices[parameterCount] = depending.getName()+" % one_dim(i) % value";
        indices[parameterCount+1] = "new_frequency";
        
        // P(depending | variable')
        
        // DEBUG
        /*
          System.out.println("GENERATOR DEBUG: About to call getFreqCode with parameters: ");
          for (int print=0; print<indices.length; print++) {
          System.out.println(indices[print]);
          }
         */
        
        ArrayList<String> toAdd=depending.getDistribution().getFreqCode(indices);
        result.addAll(toAdd);
        
        for (int i=0; i<parameterCount; i++) {
            indices[i] = indices[i].replace("% new_value", "% value");
        }
        indices[parameterCount+1] = "frequency";
        
        // P(depending | variable)
        
        // DEBUG
        /*
          System.out.println("GENERATOR DEBUG: About to call getFreqCode with parameters: ");
          for (int print=0; print<indices.length; print++) {
          System.out.println(indices[print]);
          }
         */
        
        toAdd = depending.getDistribution().getFreqCode(indices);
        result.addAll(toAdd);
        
        result.add("p_acc = p_acc * new_frequency / frequency");
        
        return result;
    }
    
    private static ArrayList<String> generateLikelihoodFormulaTwoDimensional(Variable depending,
            Set<Variable> newToBeUsed, boolean ij) throws IllegalParametersException,
            InvalidModelException, MissingFunctionException  {
        
        //System.out.println("GENERATOR DEBUG: generateLikelihoodFormulaTwoDimensional");
        ArrayList<String> result = new ArrayList<String>();
        String[] indices;
        int parameterCount = depending.getDistribution().getNumberOfParameters();
        indices = new String[parameterCount+2];
        
        for (int i=0; i<parameterCount; i++) {
            Object parameter = depending.getDistribution().getParameter(i);
            
            if (parameter instanceof Integer) {
                indices[i] = ((Integer)parameter).toString();
                continue;
            } else if (parameter instanceof Double) {
                // we must add _dp to the real constants
                indices[i] = ((Double)parameter).toString()+"_dp";
                continue;
            }
            // now we know that the parameter is a Variable.
            
            Variable varParameter=(Variable)parameter;
            
            boolean useNew = false;
            
            if (newToBeUsed.contains(varParameter)) {
                useNew = true;
            }
            indices[i] = varParameter.getName();
            if (varParameter.getEntity() == null) {
                if (useNew)
                    indices[i] += " % one_dim(1) % new_value";
                else
                    indices[i] += " % one_dim(1) % value";
            } else if (varParameter.getEntity().equals(depending.getEntity().getYCoordinate())) {
                if (ij) {
                    indices[i] += " % one_dim(i)";
                } else indices[i] += " % one_dim(j)";
                if (useNew)
                    indices[i] += " % new_value";
                else
                    indices[i] += " % value";
            } else if (varParameter.getEntity().equals(depending.getEntity().getXCoordinate())) {
                if (ij) {
                    indices[i] += " % one_dim(j)";
                } else indices[i] += " % one_dim(i)";
                if (useNew)
                    indices[i] += " % new_value";
                else
                    indices[i] += " % value";
            } else if (varParameter.getEntity().equals(depending.getEntity())) {
                if (useNew && ij)
                    indices[i] += " % two_dim(i, j) % new_value";
                else if (useNew && !ij)
                    indices[i] += " % two_dim(j, i) % new_value";
                else if (!useNew && ij)
                    indices[i] +=" % two_dim(i, j) % value";
                else
                    indices[i] +=" % two_dim(j, i) % value";
            }
        }
        if (ij) {
            indices[parameterCount] = depending.getName()+" % two_dim(i, j) % value";
        } else indices[parameterCount] = depending.getName()+" % two_dim(j, i) % value";
        indices[parameterCount+1] = "new_frequency";
        
        // P(depending | variable')
        
        // DEBUG
        /*
          System.out.println("GENERATOR DEBUG: About to call getFreqCode with parameters: ");
          for (int print=0; print<indices.length; print++) {
          System.out.println(indices[print]);
          }
         */
        
        ArrayList<String> toAdd = depending.getDistribution().getFreqCode(indices);
        result.addAll(toAdd);
        
        // DEBUG
        /*
        System.out.println("GENERATOR DEBUG: indices is (before):");
        for (int i=0; i<parameterCount; i++) {
            System.out.println(indices[i]);
        }
         */
        
        for (int i=0; i<parameterCount; i++) {
            indices[i] = indices[i].replace("% new_value", "% value");
        }
        indices[parameterCount+1] = "frequency";
        
        // DEBUG
        /*
        System.out.println("GENERATOR DEBUG: indices is (after ):");
        for (int i=0; i<parameterCount; i++) {
            System.out.println(indices[i]);
        }
         */
        
        // P(depending | variable)
        
        // DEBUG
        /*
          System.out.println("GENERATOR DEBUG: About to call getFreqCode with parameters: ");
          for (int print=0; print<indices.length; print++) {
          System.out.println(indices[print]);
          }
         */
        
        toAdd = depending.getDistribution().getFreqCode(indices);
        result.addAll(toAdd);
        
        result.add("p_acc = p_acc * new_frequency / frequency");
        
        return result;
    }
    
    private static ArrayList<String> generateTransitionFormula(Variable variable)
    throws IllegalParametersException, InvalidModelException, MissingFunctionException {
        // DEBUG
        /*
        System.out.println("GENERATOR DEBUG: Phase 3: q( " + variable.getName() + "', " + variable.getName() +
            ") / q(" + variable.getName() + ", " + variable.getName() + "')");
         */
        
        ArrayList<String> result = new ArrayList<String>();
        result.add("");
        result.add("! q(" + variable.getName() + "', " + variable.getName() +
                ") / q(" + variable.getName() + ", " + variable.getName() + "')");
        result.add("");
        
        String indexing = "";
        if (variable.getEntity() == null) {
            indexing = "one_dim(1)";
        } else if (variable.getEntity().isMatrix() == false){
            indexing = "one_dim(i)";
        } else indexing = "two_dim(i, j)";
        
        
        ArrayList<String> toAdd = new ArrayList<String>();
        
        if (variable.getStrategy() == "random walk") {
            // random walk proposal strategy
            
            if (variable.isInteger()) {
                result.add("transition_int = " + variable.getName() + " % " + indexing +
                        " % new_value - " + variable.getName() + " % " + indexing + " % value");
            } else {
                result.add("transition_real = " + variable.getName() + " % " + indexing +
                        " % new_value - " + variable.getName() + " % " + indexing +" % value");
            }
            
            int parameterCount = variable.getProposal().getNumberOfParameters();
            String[] indices = new String[parameterCount+2];
            
            for (int i=0; i<parameterCount; i++) {
                Object parameter = variable.getProposal().getParameter(i);
                
                if (parameter instanceof Integer) {
                    indices[i] = ((Integer)parameter).toString();
                    continue;
                } else if (parameter instanceof Double) {
                    // we must add _dp to the real constants
                    indices[i] = ((Double)parameter).toString()+"_dp";
                    continue;
                } else {
                    throw new InvalidModelException("Error in the proposal distribution of parameter " +
                            variable.getName() + ". Only constants can appear in proposal distributions.");
                }
            }
            
            if (variable.isInteger()) {
                indices[parameterCount] = "-transition_int";
            } else {
                indices[parameterCount] = "-transition_real";
            }
            indices[parameterCount+1] = "frequency";
            
            // DEBUG
            /*
            System.out.println("GENERATOR DEBUG: About to call getFreqCode with parameters: ");
            for (int print=0; print<indices.length; print++) {
                System.out.println(indices[print]);
            }
             */
            
            toAdd = variable.getProposal().getFreqCode(indices);
            result.addAll(toAdd);
            
            if (variable.isInteger()) {
                indices[parameterCount] = "transition_int";
            } else {
                indices[parameterCount] = "transition_real";
            }
            indices[parameterCount+1] = "new_frequency";
            
            
            // DEBUG
                /*
                  System.out.println("GENERATOR DEBUG: About to call getFreqCode with parameters: ");
                  for (int print=0; print<indices.length; print++)
                  {
                  System.out.println(indices[print]);
                  }
                 */
            
            toAdd = variable.getProposal().getFreqCode(indices);
            result.addAll(toAdd);
            
            result.add("p_acc = p_acc * frequency / new_frequency");
        } else {
            // fixed proposal strategy
            
            int parameterCount = variable.getProposal().getNumberOfParameters();
            String[] indices = new String[parameterCount+2];
            
            for (int i=0; i<parameterCount; i++) {
                Object parameter = variable.getProposal().getParameter(i);
                
                if (parameter instanceof Integer) {
                    indices[i] = ((Integer)parameter).toString();
                    continue;
                } else if (parameter instanceof Double) {
                    // we must add _dp to the real constants
                    indices[i] = ((Double)parameter).toString()+"_dp";
                    continue;
                }
            }
            
            indices[parameterCount] = variable.getName() + " % " + indexing +" % value";
            
            indices[parameterCount+1] = "frequency";
            
            // DEBUG
            /*
                System.out.println("GENERATOR DEBUG: About to call getFreqCode with parameters: ");
                for (int print=0; print<indices.length; print++) {
                    System.out.println(indices[print]);
                }
             */
            
            toAdd = variable.getProposal().getFreqCode(indices);
            result.addAll(toAdd);
            indices[parameterCount] = variable.getName() + " % " + indexing +" % new_value";
            indices[parameterCount+1] = "new_frequency";
            
            // DEBUG
            /*
                System.out.println("GENERATOR DEBUG: About to call getFreqCode with parameters: ");
                for (int print=0; print<indices.length; print++) {
                    System.out.println(indices[print]);
                }
             */
            
            toAdd = variable.getProposal().getFreqCode(indices);
            result.addAll(toAdd);
            
            result.add("p_acc = p_acc * frequency / new_frequency");
        }
        
        return result;
    }
}

